/*##COPYRIGHT
 ***********************************************************************
 *       Copyright (C) 2003, by iAnywhere Solutions, Inc.              *
 *       All rights reserved. iAnywhere Solutions is a trademark       *
 *       of Sybase, Inc.  No part of this software may be              *
 *       reproduced in any form or by any means - graphic,             *
 *       electronic or mechanical, including photocopying,             *
 *       recording, taping or information storage and retrieval        *
 *       systems - except with the written permission of               *
 *       iAnywhere Solutions, Inc.                                     *
 *                                                                     *
 ***********************************************************************
#*/

package ianywhere.demo;

/**
 * Created by iAnywhere Solutions
 * User: huojun
 * Date: 8-May-2003
 * Time: 2:22:18 PM
 */
import org.apache.xpath.XPathAPI;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import java.util.ArrayList;
import java.util.List;

/**
 * The class to parse and hold SOAP responses
 * from ASA SOAP service. ASA SOAP service
 * responses have a common format, which includes
 * SQLCODE, RESULTSET and META-DATA. Please
 * reference ASA document for detail information.
 * This class is designed for APACHE SOAP API, which
 * means Java clients have to use APACHE SOAP API to
 * access ASA SOAP service. Otherwise, users need to
 * develop their own solution to handle the SOAP responses
 * from ASA.
 */
public class SQLRowSet {
    private String _sqlCode;
    private List _columns = new ArrayList();
    private List _data = new ArrayList();
    private static final char NEW_LINE = '\n';

    /**
     * default constructor
     */
    public SQLRowSet() {
    }

    /**
     * Parse the sqlcode portion in the response
     * @param sqlCodeNode is the DOM Node passed by
     * APACHE SOAP API
     * @throws Exception
     */
    public void setSqlCode(Node sqlCodeNode) throws Exception {
        String sqlCode = getSQLCode(sqlCodeNode);
        setSqlCode(sqlCode);
    }

    /**
     * Parse the meta data and resultset
     * @param rowSetNode is the DOM Node passed by
     * APACHE SOAP API
     * @throws Exception
     */
    public void setRowSet(Node rowSetNode) throws Exception {
        Node rows = getRows(rowSetNode);
        nodeToBuffer(rows);
        setMetaData(getMetaData(rowSetNode));
    }

    /**
     * Get the number of columns in the resultset
     * @return the number of columns
     */
    public int getColumnCount() {
        return _columns.size();
    }

    /**
     * Get the SQLCODE from the database
     * @return the SQLCODE value in string
     */
    public String getSqlCode() {
        return _sqlCode;
    }

    /**
     * Get the column name based on a given column id
     * @param id is the column id
     * @return column name, if the id is invalid,
     * null is returned
     */
    public String getColumnName(int id) {
        Column column = getColumn(id);
        return null == column ? null : column._name;
    }

    /**
     * Get the column data type
     * @param id is the column id
     * @return the column data type in XML schema data types,
     * for example xsd:string
     */
    public String getColumnType(int id) {
        Column column = getColumn(id);
        return null == column ? null : column._type;
    }

    /**
     * Get the column data type
     * @param columnName is the column name
     * @return the column data type in XML schema data types,
     * for example xsd:string
     */
    public String getColumnType(String columnName) {
        Column column = getColumn(columnName);
        return null == column ? null : column._type;
    }

    /**
     * Check if a given column can be null
     * @param id is the column id
     * @return true if the column value can be null
     */
    public boolean isNullable(int id) {
        Column column = getColumn(id);
        return null == column ? false : column._nullable;
    }

    /**
     * Check if a given column can be null
     * @param columnName is the column name
     * @return true if the column value can be null
     */
    public boolean isNullable(String columnName) {
        Column column = getColumn(columnName);
        return null == column ? false : column._nullable;
    }

    /**
     * Get the a value from the resultset
     * @param row is the row number
     * @param column is the column number
     * @return the value in string
     */
    public String getValue(int row, int column) {
        List rowBuf = (List) _data.get(row);
        return (String)rowBuf.get(column);
    }

    /**
     * Get a value from the resutlset
     * @param row is the row number
     * @param columnName is the column name
     * @return the value in string
     */
    public String getValue(int row, String columnName) {
        Column column = getColumn(columnName);
        return null == column ? null : getValue(row, column._id);
    }

    /**
     * Get the column id based on a given column name
     * @param columnName is a given column name
     * @return the column id, -1 is returned if the column name
     * is invalid
     */
    public int getColumnId(String columnName) {
        Column column = getColumn(columnName);
        return null == column ? -1 : column._id;
    }

    /**
     * getthe row count of the resultset
     * @return the number of rows
     */
    public int getRowCount() {
        return _data.size();
    }

    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("META-DATA");
        sb.append(NEW_LINE);
        for (int i = 0; i < _columns.size(); i++) {
            Column column = (Column) _columns.get(i);
            sb.append(column.toString());
            sb.append(NEW_LINE);
        }
        sb.append("SQLCODE: ");
        sb.append(getSqlCode());
        sb.append(NEW_LINE);
        sb.append("RESULTSET:");
        sb.append(NEW_LINE);
        int rowCount = getRowCount();
        int columnCount = getColumnCount();
        for(int i=0; i<rowCount; i++) {
            for(int c=0; c<columnCount; c++) {
                String value = getValue(i, c);
                sb.append(getColumnName(c));
                sb.append("=");
                sb.append( null == value ? "NULL" : value );
                if( c+1 < columnCount ) {
                    sb.append(", ");
                }
            }
            sb.append(NEW_LINE);
        }
        return sb.toString();
    }

    /**
     * Set sqlcode
     * @param sqlCode
     */
    private void setSqlCode(String sqlCode) {
        _sqlCode = sqlCode;
    }

    /**
     * @param rows is the rowset node under diffgram
     */
    private void nodeToBuffer(Node rows) throws Exception {
        List rowBuffer;
        _data.clear();
        NodeList nodeList = XPathAPI.selectNodeList(rows, "//*/rowset/row");
        for (int r = 0; r < nodeList.getLength(); r++) {
            Node row = nodeList.item(r);
            NodeList columns = row.getChildNodes();
            rowBuffer = new ArrayList();
            for (int c = 0; c < columns.getLength(); c++) {
                Node node = columns.item(c);
                if (node.getNodeType() != Node.TEXT_NODE) {
                    Node firstChild = node.getFirstChild();
                    if (null != firstChild) {
                        rowBuffer.add(firstChild.getNodeValue());
                    } else {
                        rowBuffer.add(null);
                    }
                }
            }
            _data.add(rowBuffer);
        }
    }

    private Node getRows(Node rowSetNode) throws Exception {
        Node node = XPathAPI.selectSingleNode(rowSetNode, "*[local-name()='diffgram']/rowset");
        if (null == node) {
            System.out.println("Can't find rowset!");
        }
        return node;
    }

    private List getMetaData(Node rowSetNode) throws Exception {
        List l = new ArrayList();
        NodeList nodeList = XPathAPI.selectNodeList(rowSetNode, "*[local-name()='schema']/*/*/*/*/*/*/*");
        if (null != nodeList) {
            for (int i = 0; i < nodeList.getLength(); i++) {
                Node node = nodeList.item(i);
                Column c = new Column();
                NamedNodeMap attributes = node.getAttributes();
                Node namedItem = attributes.getNamedItem("name");
                c._name = namedItem.getNodeValue();
                namedItem = attributes.getNamedItem("type");
                c._type = namedItem.getNodeValue();
                c._id = i;
                namedItem = attributes.getNamedItem("minOccurs");
                c._nullable = "0".equals(namedItem.getNodeValue()) ? true : false;
                l.add(c);
            }
        }
        return l;
    }

    private String getSQLCode(Node sqlCodeNode) throws Exception {
        Node sqlCode = XPathAPI.selectSingleNode(sqlCodeNode, "//*[local-name()='sqlcode']/text()");
        return sqlCode.getNodeValue();
    }

    private void setMetaData(List metaData) {
        _columns = metaData;
    }

    private Column getColumn(String name) {
        for (int i = 0; i < _columns.size(); i++) {
            Column column = (Column) _columns.get(i);
            if (column._name.equalsIgnoreCase(name)) {
                return column;
            }
        }
        return null;
    }

    private Column getColumn(int id) {
        if (id >= 0 && id < _columns.size()) {
            Column column = (Column) _columns.get(id);
            return column;
        }
        return null;
    }

    /**
     * class to hold meta information of a column
     * in the resultset
     */
    private static class Column {
        private String _type;
        private boolean _nullable;
        private String _name;
        private int _id;

        public String getType() {
            return _type;
        }

        public void setType(String type) {
            _type = type;
        }

        public boolean isNullable() {
            return _nullable;
        }

        public void setNullable(boolean nullable) {
            _nullable = nullable;
        }

        public String getName() {
            return _name;
        }

        public void setName(String name) {
            _name = name;
        }

        public int getId() {
            return _id;
        }

        public void setId(int id) {
            _id = id;
        }

        public String toString() {
            StringBuffer sb = new StringBuffer();
            sb.append(_id);
            sb.append(':');
            sb.append(_name);
            sb.append('(');
            sb.append(_type);
            sb.append(')');
            if (!_nullable) {
                sb.append(" NOT NULL");
            }
            return sb.toString();
        }
    }
}
